;;
;; Takt editing mode for Emacs
;;
;;  Copyright (C) 1997, 2014-2016   Satoshi Nishimura
;;
;;  This program is free software: you can redistribute it and/or modify
;;  it under the terms of the GNU General Public License as published by
;;  the Free Software Foundation, either version 2 of the License, or
;;  (at your option) any later version.
;;
;;  This program is distributed in the hope that it will be useful,
;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;  GNU General Public License for more details.
;;
;;  You should have received a copy of the GNU General Public License
;;  along with this program.  If not, see <http://www.gnu.org/licenses/>.
;;

;; To use this mode, insert (something like) the following two lines 
;; to your ".emacs" file.
;;  (autoload 'takt-mode "takt-mode")
;;  (setq auto-mode-alist (append '(("\\.takt$" . takt-mode)) auto-mode-alist))

(require 'comint)
(require 'font-lock)

(defgroup takt nil
  "Major mode for editing Takt code."
  :prefix "takt-"
  :group 'languages)

(defcustom takt-temp-directory temporary-file-directory
  "Directory where temporary files are created."
  :type 'string :group 'takt)
(defcustom takt-command "takt"
  "Command path for the interpreter."
  :type 'string :group 'takt)
(defcustom takt-output-device 0
  "*Device number of the output MIDI device"
  :type 'integer :group 'takt)
(defcustom takt-command-args nil
  "Arguments given to the interpreter (list of strings)."
  :type '(repeat string) :group 'takt)

(defcustom takt-note-names 
  ["C" "C#" "D" "Eb" "E" "F" "F#" "G" "Ab" "A" "Bb" "B"]
  "Vector of note names used in step recording."
  :type `(vector ,@(make-list 12 'string)) :group 'takt)
(defcustom takt-show-velocity-in-steprec nil 
  "*If t then velocity is appended to each note during step recording."
  :type 'boolean :group 'takt)
(defcustom takt-csmove-track-on-play-file 2
  "Initial automatic cursor-moving track when takt-play-file is executed.
If nil, the cursor is not moved."
  :type 'integer :group 'takt)
(defcustom takt-set-csmove-on-play-region t
  "If t, the takt-play-region-* or takt-play-from-here-* commands 
enables the cursor-moving feature on the track at the point's position."
  :type 'boolean :group 'takt)
(defcustom takt-highlight t
  "*If t, lines containing notes at current playing position are highlighted."
  :type 'boolean :group 'takt)
(defcustom takt-indent-width 2
  "*Indent width per brace level."
  :type 'integer :group 'takt)
(defcustom takt-tab-always-indent t
  "*Non-nil means TAB in Takt mode should always reindent the current line,
regardless of where in the line point is when the TAB command is used."
  :type 'boolean :group 'takt)
(defcustom takt-frame-height 25
  "*Height of X-window frames generated by the `o' command 
in the track-summary buffer."
  :type 'integer :group 'takt)
(defcustom takt-colors-for-dark-background
  '("#804010" "red3" "orange3" "yellow3" "green3"
    "blue3" "magenta3" "gray60" "pink3" "cyan3")
  "List of highlighting colors for each track used when the window's 
background color is black (used wrap-aroundly)."
  :type '(repeat string) :group 'takt)
(defcustom takt-colors-for-light-background
  '("brown3" "red1" "orange1" "yellow1" "green1"
    "blue1" "magenta1" "gray50" "pink" "cyan1")
  "List of highlighting colors for each track used when the window's 
background color is not black (used wrap-aroundly)."
  :type '(repeat string) :group 'takt)

(defvar takt-mode-map (make-sparse-keymap) 
  "Keymap for Takt mode.")

(defvar takt-interp-buffer nil)
(defconst takt-interp-timeout 5)
(defvar takt-status 'stopped)
(defvar takt-temp-file nil
  "File name (w/o extension) of temporary file.")
(defvar takt-play-region-buffer nil
  "Dotted pair of the buffer executing takt-play-region and the full path name of the temprary file.")
(defvar takt-muted-tracks nil
  "List of track numbers for muted tracks. Buffer local.")
(defvar takt-note-list nil)
(defvar takt-nfaces nil
  "Number of available colors.  This is 1 for non-color display, 
and nil when overlays are not supported.")
(defconst takt-max-overlays 64)
(defvar takt-overlays (make-vector takt-max-overlays nil))
(defvar takt-filename-alist nil
  "List of the pairs of file ID and its associated buffer.")
(defvar takt-last-play-cmd nil)
(defvar takt-window-csmvtrack-alist nil
  "List of the pairs of Emacs window object and track number for
automatic cursor moving.  Buffer local.")
(defvar takt-track-names-alist nil
  "List of the pairs of track number and its name.")
(defvar takt-track-buffer nil)
(defvar takt-track-associated-buffer nil)
(defvar takt-track-overlay-list nil)

(defvar takt-emacs-version 
  (progn
    (string-match "\\([0-9]+\\).\\([0-9]+\\)" emacs-version)
    (defvar takt-emacs-minor-version 
      (string-to-number (substring emacs-version 
				   (match-beginning 2) (match-end 2))))
    (string-to-number (substring emacs-version 
				 (match-beginning 1) (match-end 1)))))

(define-key takt-mode-map "\C-c\C-s" 'takt-step-record)
(define-key takt-mode-map "\C-cs" 'takt-step-record)
(define-key takt-mode-map "\C-c\C-f" 'takt-play-file)
(define-key takt-mode-map "\C-cf" 'takt-play-file)
(define-key takt-mode-map "\C-c\C-q" 'takt-stop)
(define-key takt-mode-map "\C-cq" 'takt-stop)
(define-key takt-mode-map "\C-c\C-y" 'takt-play-region-all)
(define-key takt-mode-map "\C-cy" 'takt-play-region-all)
(define-key takt-mode-map "\C-c\C-u\C-y" 'takt-play-region-solo)
(define-key takt-mode-map "\C-c\C-uy" 'takt-play-region-solo)
(define-key takt-mode-map "\C-c\C-p" 'takt-play-from-here-all)
(define-key takt-mode-map "\C-cp" 'takt-play-from-here-all)
(define-key takt-mode-map "\C-c\C-u\C-p" 'takt-play-from-here-solo)
(define-key takt-mode-map "\C-c\C-up" 'takt-play-from-here-solo)
(define-key takt-mode-map "\C-c\C-l" 'takt-play-again)
(define-key takt-mode-map "\C-cl" 'takt-play-again)
(define-key takt-mode-map "\C-c\C-t" 'takt-show-track-summary)
(define-key takt-mode-map "\C-ct" 'takt-show-track-summary)
(define-key takt-mode-map "\C-c\C-d" 'takt-set-csmove-track)
(define-key takt-mode-map "\C-cd" 'takt-set-csmove-track)
(define-key takt-mode-map "\C-c\C-o" 'takt-open-track-frame)
(define-key takt-mode-map "\C-co" 'takt-open-track-frame)
(define-key takt-mode-map "\C-c\C-i" 'takt-show-interp-console)
(define-key takt-mode-map "\C-ci" 'takt-show-interp-console)
(define-key takt-mode-map "\t" 'takt-indent-command)
(define-key takt-mode-map "{" 'electric-takt-paren)
(define-key takt-mode-map "}" 'electric-takt-paren)
(define-key takt-mode-map "[" 'electric-takt-paren)
(define-key takt-mode-map "]" 'electric-takt-paren)
(define-key takt-mode-map ")" 'electric-takt-paren)

(condition-case ()
    (progn
      (define-key takt-mode-map [menu-bar] (make-sparse-keymap))
      (define-key takt-mode-map [menu-bar takt]
	(cons "Takt" (make-sparse-keymap "Takt")))
      
      (define-key takt-mode-map [menu-bar takt takt-show-interp-console]
	'("Show Console" . takt-show-interp-console))
      (define-key takt-mode-map [menu-bar takt takt-show-track-summary]
	'("Show Track Summary" . takt-show-track-summary))
      (define-key takt-mode-map [menu-bar takt takt-set-csmove-track]
	'("Set/Unset Auto Cursor Moving" . takt-set-csmove-track))
      (define-key takt-mode-map [menu-bar takt takt-open-track-frame]
	'("Open Track in Other Frame" . takt-open-track-frame))
      (define-key takt-mode-map [menu-bar takt takt-step-record]
	'("Step Record" . takt-step-record))
      (define-key takt-mode-map [menu-bar takt takt-play-again]
	'("Repeat Last Play Command" . takt-play-again))
      (define-key takt-mode-map [menu-bar takt takt-play-region-all]
	'("Play Region" . takt-play-region-all))
      (define-key takt-mode-map [menu-bar takt takt-play-region-solo]
	'("Play Region (Solo)" . takt-play-region-solo))
      (define-key takt-mode-map [menu-bar takt takt-play-from-here-all]
	'("Play from Here" . takt-play-from-here-all))
      (define-key takt-mode-map [menu-bar takt takt-play-from-here-solo]
	'("Play from Here (Solo)" . takt-play-from-here-solo))
      (define-key takt-mode-map [menu-bar takt takt-play-file]
	'("Play from Beginning" . takt-play-file))
      (define-key takt-mode-map [menu-bar takt takt-stop]
	'("Stop" . takt-stop)))
  (error nil))

(defvar takt-mode-syntax-table nil
  "Syntax table used in Takt-mode buffers.")
(if takt-mode-syntax-table
    nil
  (setq takt-mode-syntax-table (make-syntax-table))
  (modify-syntax-entry ?# "_" takt-mode-syntax-table)
  (modify-syntax-entry ?\' "\"" takt-mode-syntax-table)
  (modify-syntax-entry ?$ "_" takt-mode-syntax-table)
  (modify-syntax-entry ?% "." takt-mode-syntax-table)
  (modify-syntax-entry ?& "." takt-mode-syntax-table)
  (modify-syntax-entry ?` "." takt-mode-syntax-table)
  (modify-syntax-entry ?+ "." takt-mode-syntax-table)
  (modify-syntax-entry ?- "." takt-mode-syntax-table)
  (modify-syntax-entry ?< "." takt-mode-syntax-table)
  (modify-syntax-entry ?= "." takt-mode-syntax-table)
  (modify-syntax-entry ?> "." takt-mode-syntax-table)
  (modify-syntax-entry ?\\ "." takt-mode-syntax-table)
  (modify-syntax-entry ?| "." takt-mode-syntax-table)
  (modify-syntax-entry ?/ ". 124b" takt-mode-syntax-table)
  (modify-syntax-entry ?\n "> b" takt-mode-syntax-table)
  (if (< takt-emacs-version 21)
      ;; In Emacs 20 or before, we have a problem on nested '/* .. */'-style
      ;; comments but we don't have any solutions.
      (modify-syntax-entry ?* ". 23" takt-mode-syntax-table)
    (modify-syntax-entry ?* ". 23n" takt-mode-syntax-table))
)

;; Keyword list for font-lock
(defconst takt-keywords 
  (eval-when-compile
    (list
     (cons 
      (concat "\\(^\\|[^\\$[:word:]]\\)\\(" 
	      (regexp-opt '("nil" "xor" "rem" "eq" "neq" "is" "if" "elsif" 
			    "else" "switch" "case" "default" "while" "for" 
			    "foreach" "break" "def" "function" "eval" "var" 
			    "return" "newclass" "include" "defpitch" 
			    "defnative" "yield"))
	      "\\)\\>")  2)
     (list
      (concat "\\(^\\|[^\\$[:word:]]\\)\\(" 
	      (regexp-opt '("int" "float" "rational" "number" "symbol" "expr" 
			    "quote" "bool" "true" "false"))
	      "\\)\\>")  2 font-lock-type-face))))

;;
;; Takt mode
;;
(defun takt-mode ()
  "Major mode for editing Takt code.
\\{takt-mode-map}"
  (interactive)
  (kill-all-local-variables)
  (use-local-map takt-mode-map)
  (setq major-mode 'takt-mode)
  (setq mode-name "Takt")
  (set-syntax-table takt-mode-syntax-table)
  (set (make-local-variable 'indent-line-function) 'takt-indent-line)
  (if (null takt-temp-file)
      (setq takt-temp-file (expand-file-name (make-temp-name "takt")
					      takt-temp-directory)))
  (make-local-variable 'takt-muted-tracks)
  (make-local-variable 'takt-window-csmvtrack-alist)
  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults '(takt-keywords nil nil ((?_ . "w")
						    (?\\ . "\\"))))
  ;; status name
  (set (make-local-variable 'mode-line-process) 'takt-status)
  ;;   using "(let ((v  .." is just for avoiding warnning at byte compilation
  (let ((v (make-local-variable 'stopped))) (set v ": Stopped"))
  (let ((v (make-local-variable 'steprec))) (set v ": Step Rec"))
  (let ((v (make-local-variable 'playing))) (set v ": Playing"))
  (takt-highlight-init)
  (make-local-variable 'buffer-quit-function)
  (setq buffer-quit-function 'takt-stop)  ;; to stop with ESC ESC ESC
  (run-hooks 'takt-mode-hook))

(defun takt-cleanup ()
  "Remove temporary files."
  (condition-case ()
      (delete-file (concat takt-temp-file ".takt"))
    (error nil)))

(defun takt-start-interp ()
  "Start the interpreter process if it's not run yet."
  (interactive)
  (if (not (and takt-interp-buffer
		(comint-check-proc takt-interp-buffer)))
      (let ((comint-args `("Takt" ,takt-command nil "-NP" "-femacs"
			   ,(format "-m%d" takt-output-device)
			   ,@takt-command-args)))
	(setq takt-interp-buffer (apply 'make-comint comint-args))
	(let ((proc (get-buffer-process takt-interp-buffer)))
	  (with-current-buffer takt-interp-buffer
	    (setq comint-prompt-regexp "^\\(> \\)+[ \t]*")
	    (set-process-filter proc 'takt-output-filter)
	    (set-process-sentinel proc 'takt-interp-sentinel)
	    (make-local-variable 'buffer-quit-function)
	    (setq buffer-quit-function 'takt-stop)
	    ;; wait for the first prompt
	    (accept-process-output proc 1))))))

(defun takt-interp-eval (stmts)
  "Evaluate Takt statements in the interpreter."
  (takt-start-interp)
  (with-current-buffer takt-interp-buffer
    (goto-char (point-max))
    (insert stmts)
    (comint-send-input)))

(defun takt-shutdown-interp ()
  "Shut down the interpreter process."
  (interactive)
  (let (proc)
    (if (and (boundp 'takt-interp-buffer)
	     takt-interp-buffer 
	     (setq proc (get-buffer-process takt-interp-buffer)))
	(progn
	  (condition-case ()
	      (progn
		(takt-interrupt-process proc)
		(sit-for 0.1) ;; without this, signal seems to be sent after eof...
		(process-send-eof proc))
	    (error nil))
	  (sit-for 0.1)
	  (let ((cnt 0))
	    (while (and (eq (process-status proc) 'run)
			(or (< cnt takt-interp-timeout)
			    (delete-process proc)
			    nil))
	      (sleep-for 0.1)
	      (setq cnt (+ cnt 0.1))))))))

(defun takt-restart-interp ()
  "Restart the interpreter process."
  (interactive)
  (takt-shutdown-interp)
  (takt-start-interp))

(defun takt-show-interp-console ()
  "Start the interpreter and switch to its buffer window."
  (interactive)
  (takt-start-interp)
  (if (eq (current-buffer) takt-interp-buffer)
      nil
    (switch-to-buffer-other-window takt-interp-buffer)
    (goto-char (point-max))))

(defun takt-step-record ()
  "Start step-recording."
  (interactive)
  (takt-stop)
  (sit-for 0.1) ;; make sure that the interrupt signal is received
  (setq takt-note-list nil)
  (takt-interp-eval "$emacsStepRecord()")
  (takt-set-status 'steprec))

(defun takt-stop ()
  "Stop playing/step-recording/recording."
  (interactive)
  (let (proc)
    (if (and takt-interp-buffer 
	     (setq proc (get-buffer-process takt-interp-buffer)))
	(takt-interrupt-process proc)))
  (takt-set-status 'stopped)
  (takt-delete-all-highlights))

(defun takt-interrupt-process (proc)
  ;; On Emacs-for-MS-Windows or Meadow, interrupt-process apparently does 
  ;; not work for Cygwin-compiled programs. 
  (if (fboundp 'w32-version) 
      (quit-process proc)
    (interrupt-process proc)))
  
(defun takt-play-file ()
  "Play the song from the beginning."
  (interactive)
  (if (null (takt-get-cursor-moving-track))
      (takt-set-csmove-track takt-csmove-track-on-play-file))
  (setq takt-last-play-cmd (list 'takt-play-file (current-buffer)))
  (takt-play-region nil nil))

(defun takt-play-region-solo ()
  "Play the song within the current region using a temporary file.
It plays only the track on which the region beginning point is located." 
  (interactive)
  (setq takt-last-play-cmd (list 'takt-play-region 
				  (copy-marker (region-beginning)) 
				  (copy-marker (region-end)) t nil))
  (eval (append takt-last-play-cmd (list takt-set-csmove-on-play-region))))

(defun takt-play-from-here-solo ()
  "Play the song from the current point using a temporary file.
It plays only the track on which the current point is located." 
  (interactive)
  (setq takt-last-play-cmd (list 'takt-play-region
				  (point-marker) nil t nil)) 
  (eval (append takt-last-play-cmd (list takt-set-csmove-on-play-region))))

(defun takt-play-region-all ()
  "Play the song within the current region using a temporary file.
It plays all tracks marked as `play'."
  (interactive)
  (setq takt-last-play-cmd (list 'takt-play-region
				  (copy-marker (region-beginning)) 
				  (copy-marker (region-end)) nil nil))
  (eval (append takt-last-play-cmd (list takt-set-csmove-on-play-region))))

(defun takt-play-from-here-all ()
  "Play the song from the current point using a temporary file.
It plays all tracks marked as `play'."
  (interactive)
  (setq takt-last-play-cmd (list 'takt-play-region
				   (point-marker) nil nil nil))
  (eval (append takt-last-play-cmd (list takt-set-csmove-on-play-region))))

(defun takt-play-again ()
  "Play the song with the same parameters as the last play command."
  (interactive)
  (or takt-last-play-cmd 
      (error "No play commands are executed so far."))
  (save-excursion
    (if (eq (car takt-last-play-cmd) 'takt-play-file)
	(progn (set-buffer (car (cdr takt-last-play-cmd)))
	       (takt-play-file))
      (set-buffer (marker-buffer (car (cdr takt-last-play-cmd))))
      (eval takt-last-play-cmd))))

(defun takt-show-track-summary ()
  "Show a summary for tracks."
  (interactive)
  (takt-track-update-buffer)
  (switch-to-buffer-other-window takt-track-buffer)
;;  (takt-track-shrink-to-fit)
)

(defun takt-interp-sentinel (process event)
  (let ((obuffer (current-buffer)))
    (unwind-protect
	(progn
	  (set-buffer (process-buffer process))
	  (takt-delete-all-highlights)
	  (message "")        ; hide "Processing..."
	  (takt-set-status 'stopped))
      (set-buffer obuffer))))

(defun takt-set-status (st)
  (setq takt-status st)
  (set-buffer-modified-p (buffer-modified-p)))

(defun takt-play-region (beg end &optional solo mode csmove)
  "Play the song between beg and end using a temporary file. `beg' and/or `end'
   are possibly nil."
  ;; If region boundary is not on a token boundary, correct it.
  (save-excursion
    (if beg
	(progn
	  (goto-char beg)
	  (or (memq (following-char) '(0 ?\n ?\  ?\t))
	      (skip-chars-backward "^ \t\n{"))
	  ;; Is it on the beginning of a new-context statement?
	  (if (looking-at 
	       ;; then move the beginning point after '{'
	       (cond ((boundp 'MULE) "\\([ \t\n]*\\(\\(\\sw\\|\\s_\\|\\Ca\\)*[ \t\n]*:\\|newtrack[ \t\n]*([^\)]*)\\)[ \t\n]*{\\)")
		     (t "\\([ \t\n]*\\(\\(\\sw\\|\\s_\\)*:[ \t\n]*\\|newtrack[ \t\n]*([^\)]*)\\)[ \t\n]*{\\)")))
	      (goto-char (match-end 1)))
	  (setq beg (point))))
    (if end
	(progn
	  (goto-char end)
	  (or (memq (following-char) '(0 ?\n ?\  ?\t))
	      (skip-chars-backward "^ \t\n"))
	  (if (= end (point))
	      ;; if the end point is just after "}" or "}&",
	      ;; move the end point before it.
	      (progn 
		(skip-chars-backward " \t\n")
		(if (= (preceding-char) ?&) (backward-char))
		(if (= (preceding-char) ?})
		    (backward-char)
		  (goto-char end))))
	  (setq end (point))
	  (if (and beg (> beg end)) (setq beg end)))))
  ;; create a temp file.
  (let ((temp-buffer (generate-new-buffer "*takt-temp*"))
	(temp-file-name (concat takt-temp-file ".takt"))
	(csmvtrk (takt-get-cursor-moving-track)))
    (unwind-protect
	(progn
	  (setq takt-play-region-buffer
		(cons (current-buffer) (expand-file-name temp-file-name)))
	  (if (null beg)
	      (write-region 1 1 temp-file-name nil 0)
	    (write-region 1 beg temp-file-name nil 0)
	    (with-current-buffer temp-buffer
	      (goto-char (point-min))
	      ; the backspaces are for preserving the original column number
	      (insert " mark(\".begin\")\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b")
	      (if (and csmove csmvtrk (/= (cdr csmvtrk) 0))
		  (insert "trackname(\".cursor\")\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b"))
	      (if solo (insert "trackname(\".play\")\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b"))
	      (write-region 1 (point) temp-file-name t 0)))
	  (if (null end)
	      (write-region (or beg 1) (1+ (buffer-size)) temp-file-name t 0)
	    (write-region (or beg 1) end temp-file-name t 0)
	    (with-current-buffer temp-buffer
	      (goto-char (point-min))
	      (insert " mark(\".end\")\b\b\b\b\b\b\b\b\b\b\b\b\b")
	      (write-region 1 (point) temp-file-name t 0))
	    (write-region end (1+ (buffer-size)) temp-file-name t 0))
	  ;; play the temp file
	  (message "Processing...")
	  (takt-restart-interp)
	  (setq takt-filename-alist nil)
	  (setq takt-track-names-alist nil)
	  (setq takt-track-associated-buffer (current-buffer))
	  (takt-track-update-if-shown)
	  (takt-update-csmvtracks-in-interp)
	  (takt-update-tklist-in-interp)
	  (takt-interp-eval (concat "$emacsPlay(\"" 
				     temp-file-name
				     "\""
				     (if takt-highlight ",true" ",false")
				     (if solo ",true" ",false")
				     (if beg ",\".begin\"" ",nil")
				     (if end ",\".end\"" ",nil")
				     ")" ))
      (progn
	(kill-buffer temp-buffer))))))

(defun takt-update-tklist-in-interp ()
  (mapc (function (lambda(tk)
		    (takt-dehighlight-line tk))) takt-muted-tracks)
  (takt-interp-eval (concat "$emacsSetMute(%["
			     (mapconcat (function 
					 (lambda (x) (int-to-string x)))
					takt-muted-tracks ",")
			     "])")))

(defun takt-open-track-frame (tk)
  "Open a frame for displaying the specified track."
  (interactive "NTrack number: ") 
  (takt-open-frame tk))

;;
;; functions for process output handling and step recording
;;

(defvar takt-process-line-buffer "")

(defun takt-output-filter (proc str)
  (setq takt-process-line-buffer (concat takt-process-line-buffer str))
  (while (or (string-match "[^\n]*\n" takt-process-line-buffer)
	     (string-match "^\\(> \\)+[ \t]*" takt-process-line-buffer))
    (let* ((line-end (match-end 0))
	   (has-error (string-match 
		       "\\(Line .*: Col \\)\\|\\(Run-time error:\\)" 
		       takt-process-line-buffer))
	   (cur-line (substring takt-process-line-buffer 0 line-end)))
      (if (string-match "#\\$\\((.*)\\)\n" cur-line)
	  (let ((msg (car (read-from-string cur-line (match-beginning 1)))))
	    (cond ((or (null msg) (atom msg)) ; garbage?
		   nil)
		  ((= (car msg) ?F)
		   (apply 'takt-register-filename (cdr msg)))
		  ((= (car msg) ?H)
		   (if (eq takt-status 'playing)
		       (apply 'takt-highlight-line (cdr msg))))
		  ((= (car msg) ?G)
		   (takt-dehighlight-line (nth 1 msg)))
		  ((= (car msg) ?C)
		   (apply 'takt-move-point (cdr msg)))
		  ((= (car msg) ?A)
		   (takt-set-csmove-track (nth 1 msg)))
		  ((= (car msg) ?T)
		   (apply 'takt-set-track-name (cdr msg))
		   (message ""))        ; hide "Processing..."
		  ((= (car msg) ?D)	; echoed note-on message?
		   (takt-insert-notes))
		  ((= (car msg) ?I)	; MIDI message?
		   (if (and (= (/ (nth 1 msg) 16) 9) 
			    (/= (nth 3 msg) 0)) ; note-on message?
		       (setq takt-note-list (cons msg takt-note-list))))
		  ((= (car msg) ?P)       ; beginning of play
		   (takt-set-status 'playing))
		  ((= (car msg) ?Z)	; end of play
		   (takt-delete-all-highlights)
		   (takt-set-status 'stopped))))
	(comint-output-filter proc cur-line))
      (if has-error (takt-show-interp-console))
      (setq takt-process-line-buffer 
	    (substring takt-process-line-buffer line-end)))))

(defun takt-insert-notes ()
  (if (and (eq takt-status 'steprec)
	   (boundp 'major-mode) 
	   (eq major-mode 'takt-mode)
	   takt-note-list)
      (let (istring)
	(setq takt-note-list 
	      (sort takt-note-list (lambda (x y) (< (nth 2 x) (nth 2 y)))))
	(setq istring 
	      (concat 
	       (if (cdr takt-note-list) "[")
	       (mapconcat (lambda (note)
			    (if takt-show-velocity-in-steprec
				(format "%s(v=%d)" 
					(takt-notenum-to-str (nth 2 note))
					(nth 3 note))
			      (takt-notenum-to-str (nth 2 note))))
			  takt-note-list " ")
	       (if (cdr takt-note-list) "]")))
	(if (memq (preceding-char) '(?  ?\t 0 ?\012 ?\( ?\[ ?\{))
	    nil
	  (setq istring (concat " " istring)))
	;;(undo-boundary)
	(insert istring)
	(save-excursion
	  (let ((cur (point)))
	    (beginning-of-line)
	    (fill-individual-paragraphs (point) cur))) 
	;;(insert " ")
	;;(undo-start)  ;; undoing will not work properly without this
	))
  (setq takt-note-list nil))

(defun takt-notenum-to-str (note-number &optional relative)
  (let ((octave (- (/ note-number 12) 1))
	(n (% note-number 12)))
    (if relative 
	(concat (cond ((> octave 4) (make-string (- octave 4) ?^))
		      ((< octave 4) (make-string (- 4 octave) ?_)))
		(aref takt-note-names n))
      (concat (aref takt-note-names n) (int-to-string octave)))))

;;
;; functions for highlighting
;;
(defun takt-highlight-init ()
  (if (< takt-emacs-version 19)
      (setq takt-nfaces nil)
    (if (not (x-display-color-p))
	(progn
	  (setq takt-nfaces 1)
	  (copy-face 'highlight 'takt-face-ntrack1))
      (let ((trk 1) colors)
	(if (eq (cdr (assq 'background-mode 
			   (frame-parameters (selected-frame))))
		'dark)
	    (setq colors takt-colors-for-dark-background)
	  (setq colors takt-colors-for-light-background))
	(setq takt-nfaces (length colors))
	(mapcar 
	 (function 
	  (lambda (x)
	    (let ((nface (intern (format "takt-face-ntrack%d" trk))))
	      (make-face nface)
	      (set-face-background nface x))
	    (setq trk (1+ trk))))
	 colors)))))
	    
(defun takt-highlight-line (trk fid line)
  (if takt-nfaces 
      (let ((olbuf (cdr (assq fid takt-filename-alist))))
	(if (and olbuf (buffer-name olbuf)) ; buffer-name ensures "live" buffer
	    (let ((cbuf (current-buffer)))
	      (unwind-protect
		  (progn
		    (set-buffer olbuf)
		    (takt-dehighlight-line trk)
		    (if t ;;(not (memq trk takt-muted-tracks))
			(save-excursion
			  (let (t1 t2 ov)
			    (goto-char (point-min))
			    (forward-line (1- line))
			    (skip-chars-forward " \t")
			    (setq t1 (point))
			    (end-of-line)
			    (setq t2 (point))
			    (setq ov (make-overlay t1 t2))
			    (aset takt-overlays trk ov)
			    (overlay-put 
			     ov 'face 
			     (intern (format 
				      "takt-face-ntrack%d"
				      (1+ (% (1- trk) takt-nfaces)))))))))
		(set-buffer cbuf)))))))
	    
(defun takt-dehighlight-line (trk)
  (if takt-nfaces
      (let ((x (aref takt-overlays trk)))
	(if x (delete-overlay x)))))

(defun takt-delete-all-highlights ()
  (if takt-nfaces
      (mapcar 
       (function 
	(lambda (x) 
	  (if x (delete-overlay x))))
       takt-overlays)))

(defun takt-register-filename (fid filename)
  (setq takt-filename-alist 
	(cons (cons fid (if (and takt-play-region-buffer
				 (string= (cdr takt-play-region-buffer) 
					  (expand-file-name filename)))
			    (car takt-play-region-buffer)
			  (get-file-buffer filename)))
	      takt-filename-alist)))

(defun takt-update-csmvtracks-in-interp ()
  (takt-interp-eval 
   (format "$emacsSetCursorMovingTracks(%%[%s])" 
	   (mapconcat (function (lambda (x) (int-to-string (cdr x))))
		      takt-window-csmvtrack-alist ","))))

;;
;; functions for automatic cursor movement (score-following cursors)
;;
(defun takt-set-csmove-track (trk)
  "Set the cursor-moving feature on the specified track."
  (interactive 
   (list
    (if (numberp current-prefix-arg)
	current-prefix-arg
      (let (buf (oldtk (takt-get-cursor-moving-track)))
	(while 
	    (not (string-match 
		  "^\\s *[0-9]+\\s *$"
		  (setq buf 
			(read-from-minibuffer
			 (if (or (null oldtk) (= (cdr oldtk) 0))
			     "Track number: "
			   (format "Change track number %d to: (0 for disabling) "
			    (cdr oldtk)))))))
	  (message "Please enter a number.")
	  (sit-for 0.5))
	(string-to-number buf)))))
  (let ((old-list takt-window-csmvtrack-alist))
    (setq takt-window-csmvtrack-alist nil)
    (mapc (function (lambda(x)
		      (if (and (window-live-p (car x))
			       (not (eq (car x) (selected-window))))
			  (setq takt-window-csmvtrack-alist 
				(cons x takt-window-csmvtrack-alist)))))
	  old-list))
  (setq takt-window-csmvtrack-alist 
	(cons (cons (selected-window) trk) takt-window-csmvtrack-alist))

  (if (memq takt-status '(playing recording))
      (takt-update-csmvtracks-in-interp))

  (let ((p (assq 'title (frame-parameters (selected-frame)))))
    (if (and p 
	     (stringp (cdr p))
	     (string-match "^Track [-0-9]+$" (cdr p)))
	(modify-frame-parameters 
	 (selected-frame)
	 (list (cons 'title
		     (if (> trk 0) 
			 (format "Track %d" trk) 
		       "Track --")))))))

(defun takt-get-cursor-moving-track ()
  (assq (selected-window) takt-window-csmvtrack-alist))
  
(defun takt-move-point (fileid track line column)
  (if (< line 1) 
      (message "No notes are played on that track (yet).")
    (let ((buf (cdr (assq fileid takt-filename-alist))))
      (if (and buf (buffer-name buf)) ; buffer-name ensures "live" buffer
	  (let ((obuffer (current-buffer)))
	    (unwind-protect 
		(progn
		  (set-buffer buf)
		  (mapcar (function 
			   (lambda (wdtrk)
			     (if (and (window-live-p (car wdtrk))
				      (eq (window-buffer (car wdtrk)) buf)
				      (= (cdr wdtrk) track))
				 (let ((owindow (selected-window)))
				   (unwind-protect
				       (progn
					 (select-window (car wdtrk))
					 (goto-char (point-min))
					 (forward-line (1- line))
					 (takt-goto-column column)
					 (skip-chars-forward "_^")
					 (or (pos-visible-in-window-p 
					      (point) (car wdtrk))
					     (recenter)))
				     (select-window owindow))))))
			  takt-window-csmvtrack-alist))
	      (set-buffer obuffer)))))))

;; The column number generated by the compiler is the number of bytes
;; in the source file from the beginning of line.  Because Emacs handles
;; non-Latin characters with a different number of bytes per character (ex. 
;; 3 bytes for Kanji), counting columns after converting the text to the 
;; original code system is necessary to find a right column position.
(defun takt-goto-column (column)
  "Moves point to `column' in the current line, where `column' is measured 
in bytes in the source file."
  (if (null enable-multibyte-characters)
      (progn 
	(beginning-of-line)
	(condition-case ()
	    (forward-char (1- column))
	  (error nil)))
    (let ((obuffer (current-buffer))
	  (fcs buffer-file-coding-system)
	  (col 0)
	  t1 t2)
      (save-excursion
	(beginning-of-line)
	(setq t1 (point))
	(end-of-line)
	(setq t2 (point))
	(if (not (multibyte-string-p (buffer-substring t1 t2)))
	    (setq col column)
	  (set-buffer (get-buffer-create " *column-calc-work*"))
	  (erase-buffer)
	  (insert-buffer-substring obuffer t1 t2)
	  (encode-coding-region (point-min) (point-max) fcs)
	  (condition-case()
	      (delete-region column (point-max))
	    (error nil))
	  (decode-coding-region (point-min) (point-max) fcs)
	  (setq col (point-max))))
      (beginning-of-line)
      (condition-case ()
	  (forward-char (1- col))
	(error nil)))))

;;
;; functions for indentation
;;
(defun takt-indent-command ()
  "Indent current line as Takt code or insert a TAB character."
  (interactive)
  (if (and (not takt-tab-always-indent)
	   (save-excursion
	     (skip-chars-backward " \t")
	     (not (bolp))))
      (insert-tab)
    (takt-indent-line)))

(defun takt-indent-line ()
  "Indent current line as Takt code."
  ;; This was originally c-indent-line in c-mode.el.
  (let ((indent (calculate-takt-indent))
	beg shift-amt
	(pos (- (point-max) (point))))
    (beginning-of-line)
    (setq beg (point))
    (cond ((eq indent nil)
	   (setq indent (current-indentation)))
	  ((eq indent t)
	   (setq indent (calculate-takt-indent-within-comment))))
    (skip-chars-forward " \t")
    (setq shift-amt (- indent (current-column)))
    (if (zerop shift-amt)
	(if (> (- (point-max) pos) (point))
	    (goto-char (- (point-max) pos)))
      (delete-region beg (point))
      (indent-to indent)
      ;; If initial point was within line's indentation,
      ;; position after the indentation.  Else stay at same point in text.
      (if (> (- (point-max) pos) (point))
	  (goto-char (- (point-max) pos))))
    shift-amt))

(defun calculate-takt-indent-within-comment (&optional after-star)
  "Return the indentation amount for line inside a block comment.
Optional arg AFTER-STAR means, if lines in the comment have a leading star,
return the indentation of the text that would follow this star."
  ;; This function is the same as calculate-c-indent-within-comment in 
  ;; c-mode.el.  As recent emacs has no c-mode.el, I supplied this.
  ;; Copyright (C) 1985, 86, 87, 92, 94, 95 Free Software Foundation, Inc.
  (let (end star-start)
    (save-excursion
      (beginning-of-line)
      (skip-chars-forward " \t")
      (setq star-start (= (following-char) ?\*))
      (skip-chars-backward " \t\n")
      (setq end (point))
      (beginning-of-line)
      (skip-chars-forward " \t")
      (if after-star
          (and (looking-at "\\*")
               (re-search-forward "\\*[ \t]*")))
      (and (re-search-forward "/\\*[ \t]*" end t)
           star-start
           (not after-star)
           (goto-char (1+ (match-beginning 0))))
      (if (and (looking-at "[ \t]*$") (= (preceding-char) ?\*))
          (1+ (current-column))
        (current-column)))))

(defun takt-beginning-of-defun ()
  "Move backward to next beginning of def, class, etc."
  (and (re-search-backward "^\\s(\\|^[^]}) \t\n].*\\s(" nil 'move 1)
       (progn (beginning-of-line) t)))

(defun calculate-takt-indent (&optional parse-start column-offset)
  "Return appropriate indentation for current line as Takt code.
Returns t if in a comment."
  (save-excursion
    (or column-offset (setq column-offset 0))
    (beginning-of-line)
    (let ((indent-point (point))
	  state 
	  openning-paren)
      (if parse-start
	  (goto-char parse-start)
	(takt-beginning-of-defun)
	(setq parse-start (point)))
      ;;(while (< (point) indent-point)
      ;;  (setq parse-start (point))
      ;;  (setq state (parse-partial-sexp (point) indent-point 0))
      ;;  (setq openning-paren (car (cdr state))))
      (setq state (parse-partial-sexp (point) indent-point))
      (setq openning-paren (car (cdr state)))
      (cond ((nth 4 state)
	    ;; Inside a comment
	     t
	     )
	    ((nth 3 state)
	     ;; In a string.  Don't change the indentation.
	     (current-indentation))
	    ((null openning-paren)
	     ;; Line is at top level.
	     column-offset)
	    (t
	     ;; Line is at an intermediate level.
	     (goto-char indent-point)
	     (skip-chars-forward " \t")
	     (if (looking-at "\\s)")
		 ;; Closing } ] )
		 (progn 
		   (takt-beginning-of-command state openning-paren parse-start)
		   (max column-offset (current-indentation)))
	       (if (memq (char-after openning-paren) '(?\( ?\[))
		   ;; Inside ( .. ) or [ .. ]
		   (progn
		     (goto-char (1+ openning-paren))
		     (skip-chars-forward " \t")
		     (if (null (eolp))
			 (current-column)
		       (goto-char openning-paren)
		       (+ (current-column) takt-indent-width)))
		 ;; Inside { .. }
		 (takt-beginning-of-command state openning-paren parse-start)
		 (let (x 
		       (idt (current-indentation)))
		   (goto-char openning-paren)
		   (forward-line)
		   (forward-comment (buffer-size))
		   (if (and (< (point) indent-point)
			    (> (setq x (current-indentation)) 0))
		       x
		     (+ (max column-offset idt) takt-indent-width))))))))))

(defun takt-beginning-of-command (state openning-paren parse-start)
  (goto-char openning-paren)
  (let (tstate)
    (while
	(progn
	  (beginning-of-line)
	  (setq tstate (parse-partial-sexp parse-start (point)))
	  (>= (car tstate) (car state)))
      (goto-char (car (cdr tstate))))))

(defun electric-takt-paren (arg)
  "Insert character and correct line's indentation."
  (interactive "P")
  (if (and (not arg)
	   (eolp)
	   (save-excursion
	     (skip-chars-backward " \t")
	     (bolp)))
      (progn
	(insert last-command-event)
	(takt-indent-line)
	(save-excursion (delete-char -1))))
  (self-insert-command (prefix-numeric-value arg)))

;;
;; Track-summary buffer
;;

(defvar takt-track-mode-map (make-sparse-keymap) 
  "Keymap for Takt track summary mode.")

(define-key takt-track-mode-map "q" 'takt-track-exit)
(define-key takt-track-mode-map "j" 'forward-line)
(define-key takt-track-mode-map "k" 'previous-line)
(define-key takt-track-mode-map "p" 'takt-track-set-play)
(define-key takt-track-mode-map "m" 'takt-track-set-mute)
(define-key takt-track-mode-map " " 'takt-track-toggle)
(define-key takt-track-mode-map "a" 'takt-track-all)
(define-key takt-track-mode-map "s" 'takt-track-solo)
(define-key takt-track-mode-map "\C-l" 'takt-track-update-buffer)
(define-key takt-track-mode-map "g" 'takt-track-goto-track)
(define-key takt-track-mode-map "o" 'takt-track-open-frame)
(define-key takt-track-mode-map "?" 'describe-mode)
(define-key takt-track-mode-map "\C-m" 'forward-line)
(define-key takt-track-mode-map "\C-c\C-f" 'takt-track-play-file)
(define-key takt-track-mode-map "\C-c\C-l" 'takt-track-play-again)
(define-key takt-track-mode-map "\C-c\C-s" 'takt-track-stop)
(define-key takt-track-mode-map "\C-c\C-q" 'takt-track-stop)

(condition-case ()
    (progn
      (define-key takt-track-mode-map [menu-bar] (make-sparse-keymap))
      (define-key takt-track-mode-map [menu-bar takt]
	(cons "Takt" (make-sparse-keymap "Takt")))
      
      (define-key takt-track-mode-map [menu-bar takt takt-track-exit]
	'("Hide Track Summary" . takt-track-exit))
      (define-key takt-track-mode-map [menu-bar takt takt-track-open-frame]
	'("Open a Frame for the Track" . takt-track-open-frame))
      (define-key takt-track-mode-map [menu-bar takt takt-track-goto-track]
	'("Make the Cursor Follow the Track" . takt-track-goto-track))
      (define-key takt-track-mode-map [menu-bar takt takt-track-solo]
	'("Play/Mute - Solo" . takt-track-solo))
      (define-key takt-track-mode-map [menu-bar takt takt-track-all]
	'("Play/Mute - All" . takt-track-all))
      (define-key takt-track-mode-map [menu-bar takt takt-track-toggle]
	'("Play/Mute - Toggle" . takt-track-toggle))
      (define-key takt-track-mode-map [menu-bar takt takt-track-play-file]
	'("Play from Beginning" . takt-track-play-file))
      (define-key takt-track-mode-map [menu-bar takt takt-track-play-again]
	'("Repeat Last Play Command" . takt-track-play-again))
      (define-key takt-track-mode-map [menu-bar takt takt-track-stop]
	'("Stop" . takt-track-stop)))
  (error nil))

(defun takt-track-mode ()
  "Major mode for displaying track infomration and track muting status.
The \"P\" sign at the begining of each line means that the track 
is to be played (i.e., not muted).
Typed letters on this buffer are interpreted as commands:

 q   -- close track-summary window.
 p   -- mark track to be played.
 m   -- mark track to be muted.
 SPC -- toggle play/mute mark.
 a   -- mark all tracks to be played.
 s   -- mark that track and track 1 to be played and mark other tracks 
        to be muted (solo).
 C-l -- update track summary.
 g   -- make the cursor follow that track
 o   -- open a frame for showing that track
 ?   -- show this help.
 C-c C-f  -- play from the beginning (same as Takt mode).
 C-c C-l  -- repeat last play command (same as Takt mode).
 C-c C-q  -- stop (same as Takt mode)."
  (interactive)
  (kill-all-local-variables)
  (use-local-map takt-track-mode-map)
  (setq major-mode 'takt-track-mode)
  (setq mode-name "Takt Tracks")
  (setq buffer-read-only t)
  (make-local-variable 'buffer-quit-function)
  (setq buffer-quit-function 'takt-stop)
  (run-hooks 'takt-track-mode-hook))

(defun takt-set-track-name (tk name)
  (let ((rec (assq tk takt-track-names-alist)))
    (cond ((null rec)
	   (setq takt-track-names-alist
		 (cons (cons tk name) takt-track-names-alist))
	   (takt-track-update-if-shown))
	  ((null (cdr rec))
	   (setcdr rec name)
	   (takt-track-update-if-shown)))))

(defun takt-track-exit ()
  (interactive)
  (condition-case ()
      (replace-buffer-in-windows "*Help*")
    (error nil))
  (if (one-window-p) 
      (switch-to-buffer (other-buffer))
    (delete-window)))

(defun takt-track-play-file ()
  (interactive)
  (switch-to-buffer-other-window takt-track-associated-buffer)
  (takt-play-file))

(defun takt-track-play-again ()
  (interactive)
  (switch-to-buffer-other-window takt-track-associated-buffer)
  (takt-play-again))

(defun takt-track-stop ()
  (interactive)
  (with-current-buffer takt-track-associated-buffer
    (takt-stop)))

(defun takt-track-toggle ()
  (interactive)
  (let ((tk (takt-track-get-trackid)))
    (with-current-buffer takt-track-associated-buffer
      (if (memq tk takt-muted-tracks)
	  (setq takt-muted-tracks (delq tk takt-muted-tracks))
	(setq takt-muted-tracks (cons tk takt-muted-tracks)))
      (takt-update-tklist-in-interp)))
  (takt-put-play-marks))

(defun takt-track-set-play ()
  (interactive)
  (let ((tk (takt-track-get-trackid)))
    (with-current-buffer takt-track-associated-buffer
      (if (memq tk takt-muted-tracks)
	  (setq takt-muted-tracks (delq tk takt-muted-tracks)))
      (takt-update-tklist-in-interp)))
  (takt-put-play-marks)
  (forward-line))

(defun takt-track-set-mute ()
  (interactive)
  (let ((tk (takt-track-get-trackid)))
    (with-current-buffer takt-track-associated-buffer
      (if (memq tk takt-muted-tracks)
	  nil
	(setq takt-muted-tracks (cons tk takt-muted-tracks)))
      (takt-update-tklist-in-interp)))
  (takt-put-play-marks)
  (forward-line))

(defun takt-track-all ()
  (interactive)
  (with-current-buffer takt-track-associated-buffer
    (setq takt-muted-tracks nil)
    (takt-update-tklist-in-interp))
  (takt-put-play-marks))

(defun takt-track-solo ()
  (interactive)
  (let ((tk (takt-track-get-trackid))
	(ntrks (takt-put-play-marks))
	(i 2))
    (with-current-buffer takt-track-associated-buffer
      (setq takt-muted-tracks nil)
      (while (<= i ntrks)
	(if (/= tk i) 
	    (setq takt-muted-tracks (cons i takt-muted-tracks)))
	(setq i (1+ i)))
      (takt-update-tklist-in-interp)))
  (takt-put-play-marks))

(defun takt-track-goto-track ()
  (interactive)
  (let ((tk (takt-track-get-trackid)))
    (if (with-current-buffer (window-buffer (next-window))
	  (and (boundp 'major-mode) (eq major-mode 'takt-mode)))
	(other-window 1)
      (switch-to-buffer-other-window takt-track-associated-buffer))
    (takt-set-csmove-track tk)))

(defun takt-track-open-frame ()
  (interactive)
  (takt-open-frame (takt-track-get-trackid)))

(defun takt-open-frame (tk)
  (let* ((titlestr (format "Track %d" tk))
	 (fr (or
	      ;; don't create a new frame if it already exists
	      (let ((flist (frame-list)) result)
		(while flist
		  (if (string= (frame-parameter (car flist) 'title) titlestr)
		      (progn (setq result (car flist)) (setq flist nil))
		    (setq flist (cdr flist))))
		result)
	      (make-frame 
	       (list (cons 'height takt-frame-height)
		     (cons 'title titlestr))))))
    (if (with-current-buffer (window-buffer (next-window))
	  (and (boundp 'major-mode) (eq major-mode 'takt-mode)))
	(set-window-buffer (frame-selected-window fr)
			   (window-buffer (next-window)))
      (set-window-buffer (frame-selected-window fr)
			 takt-track-associated-buffer))
    (select-frame-set-input-focus fr)
    (let ((owindow (selected-window)))
      (unwind-protect (progn (select-window (frame-selected-window fr))
			     (takt-set-csmove-track tk))
	(select-window owindow)))))

(defun takt-track-get-trackid ()
  (beginning-of-line)
  (if (looking-at ". *\\([0-9]+\\)")
      (string-to-number 
       (buffer-substring (match-beginning 1) (match-end 1)))
    (error "Cursor not on a track.")))

(defun takt-track-update-buffer ()
  "Update the contents of track information in the track-summary buffer."
  (interactive)
  (save-excursion
    (if (null (get-buffer "*takt-track*"))
	(progn
	  (setq takt-track-buffer (generate-new-buffer "*takt-track*"))
	  (set-buffer takt-track-buffer)
	  (takt-track-mode)))
    (set-buffer takt-track-buffer)
    (if (null takt-track-associated-buffer)
	(let ((buffer-read-only nil))
	  (erase-buffer)
	  (insert "Track summary is currently unavailable.\n")
	  (insert "Try to play the file and and then type C-l on this window\n"))
      (progn
	(let ((buffer-read-only nil))
	  (erase-buffer)
	  (insert (concat "File: " 
			  (buffer-file-name takt-track-associated-buffer) 
			  "  Type ? for help.\n"))
	  (insert "  Trk   Track/Sequence Name\n")
	  (insert "= ==  ====================================================\n")
	  (if takt-track-names-alist
	      (let ((i 1) 
		    (maxtrk (apply 'max (mapcar (function 
						 (lambda (x) (car x)))
						takt-track-names-alist)))
		    trk-name)
		(while (<= i maxtrk)
		  (setq trk-name (assq i takt-track-names-alist))
		  (insert (format 
			   "  %2d  %s\n" i (or (cdr trk-name) "")))
		  (setq i (1+ i)))))
	  (takt-put-play-marks)
	  (takt-track-show-color-bars))))))

(defun takt-track-update-if-shown ()
  (if (and takt-track-buffer (get-buffer-window takt-track-buffer t))
      (takt-track-update-buffer)))

(defun takt-put-play-marks ()
  "Insert play-track marks on the beginning of each line.
Returns the number of tracks."
  (save-excursion  ; to save the point location
    (let ((buffer-read-only nil)
	  (tk 1)
	  trks)
      (with-current-buffer 
	  takt-track-associated-buffer ; to get buffer-local takt-muted-tracks
	(setq trks takt-muted-tracks))
      (goto-char 1)
      (forward-line 3)
      (while (looking-at ". *[0-9]")
	(delete-char 1)
	(insert-char (if (memq tk trks) ?\  ?P) 1)
	(forward-line 1)
	(setq tk (1+ tk)))
      (1- tk))))

;;(defun takt-track-shrink-to-fit ()
;;  "Shrink the size of track-summary window so that 
;;no blank lines are shown."
;;  (let ((owindow (selected-window))
;;	lines wh)
;;    (save-excursion
;;      (set-buffer takt-track-buffer)
;;      (setq lines (count-lines (point-min) (point-max))))
;;    (unwind-protect
;;	(progn 
;;	  (select-window (get-buffer-window takt-track-buffer))
;;	  (setq wh (1- (window-height)))
;;	  (if (< lines wh)
;;	      (shrink-window (- wh lines))))
;;      (select-window owindow))))

(defun takt-track-show-color-bars ()
  (if takt-nfaces
      (let ((tk 0) (ovlist))
	(takt-track-delete-color-bars)
	(save-excursion
	  (goto-char 1)
	  (forward-line 3)
	  (while (looking-at ". *[0-9]")
	    (forward-char 2)
	    (setq ovlist (cons (make-overlay (point) (+ (point) 2)) ovlist))
	    (overlay-put (car ovlist) 'face (intern 
					     (format "takt-face-ntrack%d"
						     (1+ (% tk takt-nfaces)))))
	    (forward-line 1)
	    (setq tk (1+ tk))))
	(setq takt-track-overlay-list ovlist))))

(defun takt-track-delete-color-bars ()
  (if takt-nfaces
      (mapc (function (lambda (x) 
			(delete-overlay x)))
	    takt-track-overlay-list))
  (setq takt-track-overlay-list nil))

;;
;; add exit hooks 
;;

(condition-case ()
    (add-hook 'kill-emacs-hook 'takt-cleanup)
  (error nil))

(condition-case ()
    (add-hook 'kill-emacs-query-functions (function (lambda ()
						      (takt-shutdown-interp)
						      t)))
  (error nil))

;;
(provide 'takt-mode)

;;
;; end of takt-mode.el
;;
